import os
import sys
import numpy as np
import pandas as pd

env_path = os.path.join(os.path.dirname(__file__), '../..')
if env_path not in sys.path:
    sys.path.append(env_path)

from pytracking.evaluation.environment import env_settings
from pytracking.evaluation import Tracker, get_dataset
from pytracking.analysis.plot_results import generate_formatted_report
davis_path = env_settings().davis_toolkit_path
if davis_path not in sys.path:
    sys.path.append(davis_path)

from davis2017.evaluation import DAVISEvaluation


def evaluate_davis(trackers, split='val'):
    """
    Evaluate a list of trackers on the davis17 dataset
    """
    settings = env_settings()
    dataset_eval = DAVISEvaluation(davis_root=settings.davis_path, task='semi-supervised', gt_set=split)

    csv_name_global = f'global_results-{split}.csv'
    csv_name_per_sequence = f'per-sequence_results-{split}.csv'

    scores = {'J&F-Mean': [], 'J-Mean': [], 'F-Mean': []}
    display_names = []
    for t in trackers:
        if t.display_name is not None:
            disp_name = t.display_name
        elif t.run_id is not None:
            disp_name = '{} {}_{:03d}'.format(t.name, t.parameter_name, t.run_id)
        else:
            disp_name = '{} {}'.format(t.name, t.parameter_name)

        display_names.append(disp_name)
        results_path = t.segmentation_dir

        # Check if the method has been evaluated before, if so read the results, otherwise compute the results
        csv_name_global_path = os.path.join(results_path, csv_name_global)
        csv_name_per_sequence_path = os.path.join(results_path, csv_name_per_sequence)
        if os.path.exists(csv_name_global_path) and os.path.exists(csv_name_per_sequence_path):
            table_g = pd.read_csv(csv_name_global_path)
            table_seq = pd.read_csv(csv_name_per_sequence_path)
        else:
            # Create dataset and evaluate
            metrics_res = dataset_eval.evaluate(results_path)
            J, F = metrics_res['J'], metrics_res['F']

            # Generate dataframe for the general results
            g_measures = ['J&F-Mean', 'J-Mean', 'J-Recall', 'J-Decay', 'F-Mean', 'F-Recall', 'F-Decay']
            final_mean = (np.mean(J["M"]) + np.mean(F["M"])) / 2.
            g_res = np.array(
                [final_mean, np.mean(J["M"]), np.mean(J["R"]), np.mean(J["D"]), np.mean(F["M"]), np.mean(F["R"]),
                 np.mean(F["D"])])
            g_res = np.reshape(g_res, [1, len(g_res)])
            table_g = pd.DataFrame(data=g_res, columns=g_measures)
            with open(csv_name_global_path, 'w') as f:
                table_g.to_csv(f, index=False, float_format="%.3f")
            # print(f'Global results saved in {csv_name_global_path}')

            # Generate a dataframe for the per sequence results
            seq_names = list(J['M_per_object'].keys())
            seq_measures = ['Sequence', 'J-Mean', 'F-Mean']
            J_per_object = [J['M_per_object'][x] for x in seq_names]
            F_per_object = [F['M_per_object'][x] for x in seq_names]
            table_seq = pd.DataFrame(data=list(zip(seq_names, J_per_object, F_per_object)), columns=seq_measures)
            with open(csv_name_per_sequence_path, 'w') as f:
                table_seq.to_csv(f, index=False, float_format="%.3f")
        scores['J&F-Mean'].append(table_g['J&F-Mean'].values[0]*100)
        scores['J-Mean'].append(table_g['J-Mean'].values[0]*100)
        scores['F-Mean'].append(table_g['F-Mean'].values[0]*100)

    report = generate_formatted_report(display_names, scores)
    print(report)

