import os
import sys
import importlib
import numpy as np
import torch
import time
import cv2 as cv
from pytracking.analysis.plot_results import get_plot_draw_styles
from pytracking.utils.load_text import load_text
from pytracking.utils.visdom import Visdom
from pytracking.evaluation import Tracker, get_dataset, trackerlist

env_path = os.path.join(os.path.dirname(__file__), '../..')
if env_path not in sys.path:
    sys.path.append(env_path)

from pytracking.evaluation.environment import env_settings


class VisdomUI:
    def __init__(self):
        self.pause_mode = True
        self.step = 0
        self.exit = False

    def callback_handler(self, data):
        if data['event_type'] == 'KeyPress':
            if data['key'] == ' ':
                self.pause_mode = not self.pause_mode
                if not self.pause_mode:
                    self.step = 1
            elif data['key'] == 'q':
                self.exit = True
            elif data['key'] == 'ArrowRight':
                if self.pause_mode:
                    self.step = 1
                else:
                    if self.step == -1:
                        self.step = 1
                    else:
                        self.step *= 2
            elif data['key'] == 'ArrowLeft':
                if self.pause_mode:
                    self.step = -1
                else:
                    if self.step == 1:
                        self.step = -1
                    else:
                        self.step *= 2

    def update(self):
        if self.pause_mode:
            self.step = 0


def read_image(image_file: str):
    im = cv.imread(image_file)
    return cv.cvtColor(im, cv.COLOR_BGR2RGB)


def playback_results(trackers, sequence, visdom_info=None):
    plot_draw_styles = get_plot_draw_styles()

    visdom_ui = VisdomUI()
    visdom = Visdom(1, {'handler': visdom_ui.callback_handler, 'win_id': 'Tracking'},
                    visdom_info=visdom_info)

    # Load results for all trackers
    tracker_results = []
    for trk_id, trk in enumerate(trackers):
        base_results_path = '{}/{}'.format(trk.results_dir, sequence.name)
        results_path = '{}.txt'.format(base_results_path)

        if os.path.isfile(results_path):
            pred_bb = torch.tensor(load_text(str(results_path), delimiter='\t', dtype=np.float64))
        else:
            raise Exception('Result not found. {}'.format(results_path))

        tracker_results.append(pred_bb)

    # Convert to list of shape seq_length * num_trackers * 4
    tracker_results = torch.stack(tracker_results, dim=1)
    tracker_names = [t.display_name for t in trackers]

    images = [read_image(s) for s in sequence.frames]
    frame_number = 0
    while not visdom_ui.exit:
        image = images[frame_number]

        # boxes = [torch.tensor(b) for b in tracker_results[frame_number].tolist()]
        boxes = tracker_results[frame_number].tolist()
        visdom.register([image] + boxes, 'Tracking', 0, title='Tracking')

        time.sleep(0.1)
        while visdom_ui.step == 0 and not visdom_ui.exit:
            time.sleep(0.01)

        frame_number = min(max(frame_number + visdom_ui.step, 0), len(sequence.frames) - 1)
        visdom_ui.update()

if __name__ == '__main__':
    # playback_results('goutam', 'caramba0', sequence_id=47, visdom_info={'use_visdom': True, 'server': '127.0.0.1', 'port': 8097})
    trackers = trackerlist('atom_cvpr19', 'default', range(0, 5), 'ATOM')
    dataset = get_dataset('otb')
    playback_results(trackers, dataset[1], visdom_info={'use_visdom': True, 'server': '127.0.0.1', 'port': 8097})